"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generic = exports.parameterProperties = exports.noop = exports.type = exports.ignore = exports.mergeDecorator = exports.decorate = exports.decorateClass = exports.decorateProperty = exports.decorateMethod = exports.decorateParameter = void 0;
require("reflect-metadata");
const helpers_1 = require("./helpers");
const types_1 = require("./types");
// --------------------------------------------------------------------- //
// ------------------------------- HELPER ------------------------------ //
// --------------------------------------------------------------------- //
function addDecorator(target, decorator) {
    const decorators = Reflect.getOwnMetadata(types_1.DECORATOR_KEY, target) || [];
    decorators.push(decorator);
    Reflect.defineMetadata(types_1.DECORATOR_KEY, decorators, target);
}
function decorateParameter(data, option) {
    return decorate(data, ["Parameter"], option);
}
exports.decorateParameter = decorateParameter;
function decorateMethod(data, option) {
    return decorate(data, ["Method"], option);
}
exports.decorateMethod = decorateMethod;
function decorateProperty(data, option) {
    return decorate(data, ["Property", "Parameter"], option);
}
exports.decorateProperty = decorateProperty;
function decorateClass(data, option) {
    return decorate(data, ["Class"], option);
}
exports.decorateClass = decorateClass;
function decorate(data, targetTypes = [], option) {
    const throwIfNotOfType = (target) => {
        if (targetTypes.length > 0 && !targetTypes.some(x => x === target))
            throw new Error(`Reflect Error: Decorator of type ${targetTypes.join(", ")} applied into ${target}`);
    };
    const opt = Object.assign({ allowMultiple: true, inherit: true, applyTo: [], removeApplied: true }, option);
    return (...args) => {
        const theData = typeof data === "function" ? data(...args) : data;
        if (!opt.allowMultiple && !theData[types_1.DecoratorId]) {
            const ctorName = helpers_1.metadata.isConstructor(args[0]) ? args[0].name : args[0].constructor.name;
            throw new Error(`Reflect Error: Decorator with allowMultiple set to false must have DecoratorId property in ${ctorName}`);
        }
        //class decorator
        if (args.length === 1) {
            throwIfNotOfType("Class");
            return addDecorator(args[0], Object.assign({ targetType: "Class", target: args[0].name, value: typeof data === "function" ? data(args[0]) : data }, opt));
        }
        //parameter decorator
        if (args.length === 3 && typeof args[2] === "number") {
            throwIfNotOfType("Parameter");
            const isCtorParam = helpers_1.metadata.isConstructor(args[0]);
            const targetType = isCtorParam ? args[0] : args[0].constructor;
            const targetName = isCtorParam ? "constructor" : args[1];
            return addDecorator(targetType, Object.assign({ targetType: "Parameter", target: targetName, targetIndex: args[2], value: typeof data === "function" ? data(targetType, targetName, args[2]) : data }, opt));
        }
        //property
        if (args[2] === undefined || args[2].get || args[2].set) {
            throwIfNotOfType("Property");
            return addDecorator(args[0].constructor, Object.assign({ targetType: "Property", target: args[1], value: typeof data === "function" ? data(args[0].constructor, args[1]) : data }, opt));
        }
        throwIfNotOfType("Method");
        return addDecorator(args[0].constructor, Object.assign({ targetType: "Method", target: args[1], value: typeof data === "function" ? data(args[0].constructor, args[1]) : data }, opt));
    };
}
exports.decorate = decorate;
function mergeDecorator(...fn) {
    return (...args) => {
        fn.forEach(x => x(...args));
    };
}
exports.mergeDecorator = mergeDecorator;
const symIgnore = Symbol("ignore");
const symOverride = Symbol("override");
const symArray = Symbol("array");
const symParamProp = Symbol("paramProp");
const symNoop = Symbol("noop");
function ignore() {
    return decorate({ [types_1.DecoratorId]: symIgnore, kind: "Ignore" }, ["Parameter", "Method", "Property"], { allowMultiple: false });
}
exports.ignore = ignore;
function type(type, ...genericParams) {
    // type is not inheritable because derived class can define their own type override
    return decorate((target) => ({ [types_1.DecoratorId]: symOverride, kind: "Override", type, genericParams, target }), ["Parameter", "Method", "Property"], { inherit: false, allowMultiple: false });
}
exports.type = type;
function noop() {
    // type is not inheritable because derived class can define their own type override
    return decorate((target) => ({ [types_1.DecoratorId]: symNoop, kind: "Noop", target }), undefined, { inherit: false, allowMultiple: false });
}
exports.noop = noop;
function parameterProperties() {
    return decorateClass({ [types_1.DecoratorId]: symParamProp, type: "ParameterProperties" }, { allowMultiple: false });
}
exports.parameterProperties = parameterProperties;
var generic;
(function (generic) {
    const symGenericType = Symbol("genericType");
    const symGenericTemplate = Symbol("genericTemplate");
    function template(...templates) {
        return decorateClass(target => ({ [types_1.DecoratorId]: symGenericTemplate, kind: "GenericTemplate", templates, target }), { inherit: false, allowMultiple: false });
    }
    generic.template = template;
    function type(...types) {
        return decorateClass(target => ({ [types_1.DecoratorId]: symGenericType, kind: "GenericType", types, target }), { inherit: false, allowMultiple: false });
    }
    generic.type = type;
    /**
     * Create generic class dynamically
     * @param parent Super class that the class inherited from
     * @param params List of generic type parameters
     */
    function create(parent, ...params) {
        const opt = (typeof parent === "object") ? parent : { parent: parent, name: "DynamicType" };
        const Type = helpers_1.createClass(Object.assign(Object.assign({}, opt), { genericParams: params }));
        return Type;
    }
    generic.create = create;
})(generic = exports.generic || (exports.generic = {}));
