"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getClassMembers = exports.getFunctionParameters = exports.getConstructorParameters = exports.getMethodParameters = exports.parseFunction = exports.parseClass = void 0;
const acorn_1 = require("acorn");
const types_1 = require("./types");
// --------------------------------------------------------------------- //
// ------------------------------ HELPERS ------------------------------ //
// --------------------------------------------------------------------- //
function getNode(node, criteria) {
    if (criteria(node))
        return node;
    if (!node.body)
        return;
    if (Array.isArray(node.body)) {
        for (const child of node.body) {
            const result = getNode(child, criteria);
            if (result)
                return result;
        }
    }
    return getNode(node.body, criteria);
}
function getNamesFromAst(nodes) {
    const getName = (node) => {
        if (node.type === "Identifier")
            return node.name;
        if (node.type === "AssignmentPattern")
            return node.left.name;
        if (node.type === "RestElement")
            return node.argument.name;
        if (node.type === "Property") {
            if (node.value.type === "Identifier")
                return node.value.name;
            else {
                const result = {};
                result[node.key.name] = getName(node.value);
                return result;
            }
        }
        //if (node.type === "ObjectPattern") {
        return node.properties.map((x) => getName(x));
        //}
    };
    return nodes.map(x => getName(x)).filter((x) => !!x);
}
function getCode(fn) {
    const code = fn.toString();
    if (code.search(/^class(\s*)extends\s*option.parent\s*{\s*}/gm) > -1) {
        return "class DynamicClass extends parent {}";
    }
    else
        return code.replace("[native code]", "");
}
function getMethodParameters(fn, method) {
    const body = getCode(fn);
    const ast = acorn_1.parse(body, { ecmaVersion: 2020 });
    const ctor = getNode(ast, x => x.type === "MethodDefinition" && x.kind === "method" && x.key.name === method);
    return getNamesFromAst(ctor ? ctor.value.params : []);
}
exports.getMethodParameters = getMethodParameters;
function getConstructorParameters(fn) {
    const body = getCode(fn);
    const ast = acorn_1.parse(body, { ecmaVersion: 2020 });
    const ctor = getNode(ast, x => x.type === "MethodDefinition" && x.kind === "constructor");
    return getNamesFromAst(ctor ? ctor.value.params : []);
}
exports.getConstructorParameters = getConstructorParameters;
function getFunctionParameters(fn) {
    try {
        const body = getCode(fn);
        const ast = acorn_1.parse(body, { ecmaVersion: 2020 });
        return getNamesFromAst(ast.body[0].params);
    }
    catch (_a) {
        return [];
    }
}
exports.getFunctionParameters = getFunctionParameters;
function getClassMembers(fun) {
    const isGetter = (name) => Object.getOwnPropertyDescriptor(fun.prototype, name).get;
    const isSetter = (name) => Object.getOwnPropertyDescriptor(fun.prototype, name).set;
    const isFunction = (name) => typeof fun.prototype[name] === "function";
    const members = Object.getOwnPropertyNames(fun.prototype)
        .filter(name => isGetter(name) || isSetter(name) || isFunction(name));
    const properties = (Reflect.getOwnMetadata(types_1.DECORATOR_KEY, fun) || [])
        .filter((x) => x.targetType === "Property")
        .map((x) => x.target);
    const names = members.concat(properties)
        .filter(name => name !== "constructor" && !~name.indexOf("__"));
    return [...new Set(names)];
}
exports.getClassMembers = getClassMembers;
function printDestruct(params) {
    const result = [];
    for (const key in params) {
        const par = params[key];
        if (typeof par === "string")
            result.push(par);
        else {
            const key = Object.keys(par)[0];
            result.push(`${key}: ${printDestruct(par[key])}`);
        }
    }
    return `{ ${result.join(", ")} }`;
}
// --------------------------------------------------------------------- //
// ------------------------------- PARSER ------------------------------ //
// --------------------------------------------------------------------- //
function parseParameter(name, index) {
    let parName;
    let fields = {};
    if (typeof name === "object") {
        parName = printDestruct(name);
        fields = name;
    }
    else {
        parName = name;
    }
    return { kind: "Parameter", name: parName, decorators: [], fields, index, type: undefined };
}
function parseFunction(fn) {
    const parameters = getFunctionParameters(fn).map((x, i) => (Object.assign(Object.assign({}, parseParameter(x, i)), { type: undefined, typeClassification: undefined })));
    return { kind: "Function", name: fn.name, parameters, returnType: undefined };
}
exports.parseFunction = parseFunction;
function parseMethods(owner) {
    const result = [];
    const members = getClassMembers(owner);
    for (const name of members) {
        const des = Reflect.getOwnPropertyDescriptor(owner.prototype, name);
        if (des && typeof des.value === "function" && !des.get && !des.set) {
            const parameters = getMethodParameters(owner, name).map((x, i) => parseParameter(x, i));
            result.push({ kind: "Method", name, parameters, decorators: [], returnType: undefined });
        }
    }
    return result;
}
function parseProperties(owner) {
    const result = [];
    const members = getClassMembers(owner);
    for (const name of members) {
        const des = Reflect.getOwnPropertyDescriptor(owner.prototype, name);
        if (!des || des.get || des.set) {
            result.push({ kind: "Property", name, decorators: [], get: des === null || des === void 0 ? void 0 : des.get, set: des === null || des === void 0 ? void 0 : des.set });
        }
    }
    // include constructor parameter
    const params = getConstructorParameters(owner);
    for (const [i, name] of params.entries()) {
        const _a = parseParameter(name, i), { fields } = _a, par = __rest(_a, ["fields"]);
        result.push(Object.assign(Object.assign({}, par), { kind: "Property", isParameter: true, get: undefined, set: undefined }));
    }
    return result;
}
function parseConstructor(fn) {
    const params = getConstructorParameters(fn);
    return {
        kind: "Constructor",
        name: "constructor",
        parameters: params.map((x, i) => parseParameter(x, i)),
    };
}
function parseClass(fn) {
    const proto = Object.getPrototypeOf(fn);
    return {
        kind: "Class", name: fn.name, type: fn, decorators: [],
        methods: parseMethods(fn),
        properties: parseProperties(fn),
        ctor: parseConstructor(fn),
        super: proto.prototype ? proto : Object,
    };
}
exports.parseClass = parseClass;
